import assert from "assert";
import fs from "fs";
import yml from "js-yaml";
import { CLOUDFORMATION_SCHEMA } from "js-yaml-cloudformation-schema";
import { parse } from "@code-like-a-carpenter/foundation-parser";
import { makePlugin } from "@code-like-a-carpenter/graphql-codegen-helpers";
import { defineEnricher, defineReactor, defineTableCdc } from "./cdc";
import { ConfigSchema } from "./config";
import { combineFragments } from "./fragments/combine-fragments";
import { defineTable } from "./table";
import { schema } from "@code-like-a-carpenter/foundation-intermediate-representation";
function getInitialTemplate({
  sourceTemplate
}) {
  if (sourceTemplate) {
    const raw = fs.readFileSync(sourceTemplate, "utf8");
    try {
      return JSON.parse(raw);
    } catch {
      return yml.load(raw, {
        schema: CLOUDFORMATION_SCHEMA
      });
    }
  }
  return {
    AWSTemplateFormatVersion: "2010-09-09",
    Resources: {},
    Transform: ["AWS::Serverless-2016-10-31", "AWS::LanguageExtensions"]
  };
}
const plugin = makePlugin(
  ConfigSchema,
  (schema2, documents, config, info) => {
    const outputFile = info?.outputFile;
    assert(outputFile, "outputFile is required");
    const { models, tables } = parse(schema2, documents, config, info);
    const allResources = combineFragments(
      ...tables.map(
        (table) => combineFragments(defineTableCdc(table, config), defineTable(table))
      ),
      ...models.flatMap(
        (model) => model.changeDataCaptureConfig.map((cdc) => {
          const { type } = cdc;
          switch (type) {
            case "ENRICHER":
              return defineEnricher(config, model, cdc);
            case "REACTOR":
              return defineReactor(config, model, cdc);
            default:
              throw new Error(`Unexpected CDC type ${type}`);
          }
        })
      )
    );
    const initialTemplate = getInitialTemplate(config);
    const tpl = {
      ...initialTemplate,
      Conditions: {
        ...initialTemplate.Conditions,
        ...allResources.Conditions
      },
      Globals: {
        Function: {
          Handler: "index.handler",
          MemorySize: 256,
          Runtime: "nodejs18.x",
          Timeout: 30,
          Tracing: "Active",
          ...initialTemplate?.Globals?.Function,
          Environment: {
            // @ts-expect-error - typedef treats `Environment` as `unknown`
            ...initialTemplate?.Globals?.Function?.Environment,
            Variables: {
              // @ts-expect-error - typedef treats `Environment` as `unknown`
              ...initialTemplate?.Globals?.Function?.Environment?.Variables,
              // @ts-expect-error - typedef treats `Environment` as `unknown`
              ...allResources?.Globals?.Function?.Environment?.Variables
            }
          }
        }
      },
      Outputs: {
        ...initialTemplate.Outputs,
        ...allResources.Outputs
      },
      Parameters: {
        ...initialTemplate.Parameters,
        ...allResources.Parameters,
        StageName: {
          AllowedValues: ["development", "production", "test"],
          Default: "development",
          Description: "The name of the stage",
          Type: "String"
        }
      },
      Resources: {
        ...initialTemplate.Resources,
        ...allResources.Resources
      }
    };
    const variables = tpl?.Globals?.Function?.Environment?.Variables ?? {};
    const tablesNames = Object.keys(variables).filter((name) => name.startsWith("TABLE_")).reduce((acc, name) => {
      acc[name] = variables[name];
      delete variables[name];
      return acc;
    }, {});
    if (Object.keys(tablesNames).length) {
      variables.FOUNDATION_TABLE_NAMES = { "Fn::ToJsonString": tablesNames };
    }
    Object.entries(tpl.Resources).filter(
      ([, resource]) => resource.Type === "AWS::Serverless::Application"
    ).forEach(([, resource]) => {
      assert(resource.Type === "AWS::Serverless::Application");
      resource.Properties = {
        ...resource.Properties,
        Parameters: {
          ...resource.Properties?.Parameters ?? {},
          TableNames: { "Fn::ToJsonString": tablesNames }
        }
      };
    });
    const { format } = config.outputConfig;
    if (format === "json") {
      return JSON.stringify(tpl, null, 2);
    }
    return yml.dump(tpl, config.outputConfig.yamlConfig);
  }
);
export {
  schema as addToSchema,
  plugin
};
